/*
 *  Support-Routinen fr Allocate/Deallocate
 *
 *  Version: $VER: AllocSupport.c 1.0  D. Ghler (11.07.94)
 */

#include "AllocSupport.h"

/*
 * AllocMemHeader alloziert einen neuen Block Speicher von
 *                der globalen Speicherliste des Betriebs-
 *                systems. Vom neuen Block knnen dann mit
 *                AllocateVec() und DeallocateVec() Blcke
 *                alloziert und wieder zurckgegeben werden.
 */

struct MemHeader *AllocMemHeader(ULONG Size, ULONG Flags)
{
   APTR             block;
   struct MemHeader *mh      = 0L;
   struct MemChunk  *mc;

   /* Speicher fr alles besorgen */
   if (block = AllocMem(Size+sizeof(*mh),Flags))
   {
      /* MemHeader in den Speicherblock legen */
      mh = (struct MemHeader *)block;

      /* block zeigt hiernach auf den eigentlichen Anfang */
      /* des freien Speicherbereichs mit der Gre Size */
      block = (UBYTE *)block + sizeof(*mh);

      /* mc beschreibt das einzige freie Stck im Block */
      mc = (struct MemChunk *) block;

      /* jetzt wird die Speicher-Struktur initialisiert */
      mh->mh_Node.ln_Type = NT_MEMORY;
      mh->mh_Node.ln_Name = 0;
      mh->mh_Node.ln_Succ = NULL;
      mh->mh_Node.ln_Pred = NULL;
      mh->mh_First        = (struct MemChunk *)block;
      mh->mh_Lower        = block;
      mh->mh_Upper        = (UBYTE *)block + Size;
      mh->mh_Free         = Size; // - sizeof(*mc);

      /* Der Bereich enthlt zu Beginn nur 1 Stck, und
         das ist genau Size Byte gro */
      mc->mc_Next         = NULL;
      mc->mc_Bytes        = mh->mh_Free;
   }

   return mh;
}

/*
 * FreeMemHeader gibt den Speicher wieder frei, der mit
 *               AllocMemheader alloziert wurde.
 */

void FreeMemHeader(struct MemHeader *mh)
{
   FreeMem(mh,((UBYTE *)mh->mh_Upper - (UBYTE *)mh->mh_Lower)
              + sizeof(*mh));
}

/*
 * AllocateVec arbeitet wie Allocate, mit dem Unterschied,
 *             da es selbst speichert, wir gro der Block
 *             war. Damit mu man beim Freigeben des Blocks
 *             die Lnge nicht angeben.
 *
 * WICHTIG:    Zum Freigeben immer DeallocateVec() benutzen,
 *             nie Deallocate(), da sonst Speicher verloren
 *             geht.
 */

void *AllocateVec(struct MemHeader *mh, ULONG len)
{
   APTR mem = 0L;
   ULONG memlen = len + sizeof(ULONG);

   if ((len + 4) <= mh->mh_Free)
   {  if (mem = Allocate(mh,memlen))
      {  *(ULONG *)mem = memlen;
         mem = (ULONG *)mem + 1;
      }
   }

   return mem;
}

/*
 * DeallocateVec ist das Gegenstck zu AllocateVec(). Es
 *               liest die Lnge aus der Langwort-Adresse
 *               vor dem freizugebenden Speicherblock und
 *               gibt den Block frei.
 *
 * WICHTIG:      Bitte diese Funktion nur mit AllocateVec()
 *               benutzen, nie mit Allocate()!
 */

void DeallocateVec(struct MemHeader *mh, APTR mem)
{
   mem = ((ULONG *)mem - 1);

   Deallocate(mh, mem, *(ULONG *)mem);
}
